
#= 

# 3D Linear Elasticity
([source code](SOURCE_URL))

This example computes the solution ``\mathbf{u}`` of the linear elasticity problem
```math
\begin{aligned}
-\mathrm{div} (\mathbb{C} \epsilon(\mathbf{u})) & = \mathbf{f} \quad \text{in } \Omega\\
\mathbb{C} \epsilon(\mathbf{u}) \cdot \mathbf{n} & = \mathbf{g} \quad \text{along } \Gamma_N
\end{aligned}
```
with exterior force ``\mathbf{f}``, Neumann boundary force ``\mathbf{g}``, and the stiffness tensor ``\mathbb{C}``
for isotropic media.
=#


module Example_3DElasticity

using GradientRobustMultiPhysics

## problem data for Neumann boundary
function neumann_force_right!(result)
    result[1] = 0
    result[2] = 0
    result[3] = 10
end    

## everything is wrapped in a main function
function main(; Plotter = nothing, verbosity = 1)

    ## mesh = scaled unit cube and 3 uniform refinements
    xgrid = uniform_refine(grid_unitcube(Tetrahedron3D; scale = [4//3,3//4,1//4]), 3)

    ## parameters for isotropic elasticity tensor
    elasticity_modulus = 1000 # elasticity modulus
    poisson_number = 1//3 # Poisson number
    shear_modulus = (1/(2+2*poisson_number))*elasticity_modulus
    lambda = poisson_number*elasticity_modulus/((1-2*poisson_number)*3*(1+poisson_number))

    ## choose finite element type
    FEType = H1P1{3} # P1-Courant

    ## negotiate data functions to the package
    user_function_neumann_bnd = DataFunction(neumann_force_right!, [3,3]; name = "g", dependencies = "", quadorder = 0)

    ## PDE description via prototype
    LinElastProblem = LinearElasticityProblem(3; shear_modulus = shear_modulus, lambda = lambda)

    ## add Neumann boundary data on right side (boundary region 3)
    add_rhsdata!(LinElastProblem, 1,  RhsOperator(Identity, [3], user_function_neumann_bnd; on_boundary = true))

    ## add Dirichlet boundary data to unknown 1 (displacement) on left side (boundary region 5)
    add_boundarydata!(LinElastProblem, 1, [5], HomogeneousDirichletBoundary)

    ## show problem definition
    show(LinElastProblem)

    ## generate FESpace
    FES = FESpace{FEType}(xgrid)

    ## solve PDE
    Solution = FEVector{Float64}("displacement",FES)
    solve!(Solution, LinElastProblem; verbosity = verbosity)

    ## plot (Plotter = Makie should work)
    GradientRobustMultiPhysics.plot(Solution, [1], [Identity]; Plotter = Plotter, verbosity = verbosity)

end

end